/* ========================================================================
 * Copyright (c) 2004-2006 The University of Washington
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ========================================================================
 */


/* auth location module
   see: doc/module.html 
 */

#include "config.h"

/* debug set to PRINTF but not PRINTF2 */
#define DEBUG
#if defined(DEBUG)
#define PRINTF  if (1) printf
#define PRINTF2 if (0) printf
#else 
#define PRINTF  if (0) printf
#define PRINTF2 if (0) printf
#endif

#define CORE_PRIVATE
#include "httpd.h"
#include "http_log.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_core.h"
/* #include "ap_alloc.h" */

#ifdef APACHE2
typedef apr_pool_t pool;
typedef apr_table_t table;
#endif


module auth_location_module;

#ifdef APACHE2

#include "apr_strings.h"

#define MY_LOG_DEBUG APLOG_MARK,APLOG_DEBUG,0
#define MY_LOG_INFO  APLOG_MARK,APLOG_INFO,0
#define MY_LOG_ERR   APLOG_MARK,APLOG_ERR,0
#define MY_LOG_EMERG APLOG_MARK,APLOG_EMERG,0
#define USER user
#define AUTH_TYPE ap_auth_type

#define ap_palloc apr_palloc
#define ap_pcalloc apr_pcalloc
#define ap_pstrcat apr_pstrcat
#define ap_pstrdup apr_pstrdup
#define ap_pstrndup apr_pstrndup
#define ap_table_add apr_table_add
#define ap_table_get apr_table_get
#define ap_table_set apr_table_set
#define ap_table_setn apr_table_setn
#define ap_table_merge apr_table_merge
#define ap_overlay_tables apr_table_overlay
typedef apr_pool_t ap_pool;
#define ap_send_http_header(r) ;

#else /* apache 1.3 */

#define MY_LOG_DEBUG APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO
#define MY_LOG_INFO  APLOG_MARK,APLOG_INFO|APLOG_NOERRNO
#define MY_LOG_ERR   APLOG_MARK,APLOG_ERR
#define MY_LOG_EMERG APLOG_MARK,APLOG_EMERG|APLOG_NOERRNO
#define USER connection->user
#define AUTH_TYPE connection->ap_auth_type
#define APR_SUCCESS HTTP_OK
#define AP_RAW_ARGS func
#define AP_TAKE1 func
#define AP_FLAG func

#endif /* which apache */


/* module definitions and structures */

/* Arg string keywords to indicate guest or not 
   on the 'permit' page.  Chooser page has:
   <a href="{GUEST_YES}"> for example. */

#define PMT_RESP_YES "AuthLocationGuest_Yes"
#define PMT_RESP_NO  "AuthLocationGuest_No"
#define PMT_PAGE_YES "{GUEST_YES}"
#define PMT_PAGE_NO "{GUEST_NO}"

#define IP_IP      0
#define IP_STAR    1
#define IP_RANGE   2

typedef struct AuthLocIpRange_ {
  struct AuthLocIpRange_ *next;
  char *ipstr;
  int type;
  int lo;
  int hi;
} AuthLocIpRange;

typedef struct AuthLocName_ {
  struct AuthLocName_ *next;
  char *name;
  AuthLocIpRange *ip;
} AuthLocName;

#define AppURL   1
#define AppQS    2

typedef struct AuthLocAppName_ {
  struct AuthLocAppName_ *next;
  char *name;
  char *str;
  int where;
} AuthLocAppName;

#define PERMIT_PMT_ANON  1
#define PERMIT_REQ_ANON  2

typedef struct AuthLocPermit_ {
  struct AuthLocPermit_ *next;
  int type;
  char *appname;
  char *locname;
  char *id;
  char *page;
} AuthLocPermit;


typedef struct {
   AuthLocName *loc;
   AuthLocAppName *app;
   AuthLocPermit *permit;
} loc_dir_config_rec;



/* ------ Application and Location tools ------- */

static char *next_word(char **ptr)
{
   char *p = *ptr;
   char *ret;
   /* PRINTF("al: word in: %s\n", *ptr); */
   while (*p && (isspace(*p)||(*p==','))) p++;
   ret = p;
   while (*p && !(isspace(*p)||(*p==','))) p++;
   if (*p) *p++ = '\0';
   *ptr = p;
   return (ret);
}

/* Parse an ip definition */

static AuthLocIpRange *parse_ip(pool *p, char *str)
{
   AuthLocIpRange *ip = (AuthLocIpRange*) malloc(sizeof(AuthLocIpRange));
   char *s, *t;
   memset(ip, '\0', sizeof(AuthLocIpRange));
   PRINTF2("al: parse ip %s\n", str);
      if (s=strchr(str,'*')) {
         *s = '\0';
         ip->type = IP_STAR;
         ip->ipstr = ap_pstrdup(p, str);
      } else if (s=strchr(str, '-')) {
         *s++ = '\0';
         if (t=strrchr(str,'.')) {
           ip->lo = atoi(t+1);
           *(t+1) = '\0';
         } else ip->lo = 0;
         ip->hi = atoi(s);
         ip->ipstr = ap_pstrdup(p, str);
         ip->type = IP_RANGE;
      } else {
         ip->ipstr = ap_pstrdup(p, str);
         ip->type = IP_IP;
      }
   PRINTF2("al: str=%s, lo=%d, hi=%d\n", ip->ipstr, ip->lo, ip->hi);
   return (ip);
}

/* Find a location by name, Possibly make new if not found */

static AuthLocName *find_loc_by_name(pool *p, loc_dir_config_rec *dir_conf,
     char *name, int new)
{
  AuthLocName *last = NULL;
  AuthLocName *l;
  PRINTF2("al: look (%d) for loc %s\n", new, name);
  for (l=dir_conf->loc; l; last=l,l=l->next) {
     if (!strcmp(l->name, name)) return (l);
  }
  PRINTF2(new? "al: adding\n": "al: not found\n");
  /* not found */
  if (!new) return (NULL);
  l = (AuthLocName*) ap_pcalloc (p, sizeof (AuthLocName));
  if (last) last->next = l;
  else dir_conf->loc = l;
  l->name = ap_pstrdup(p, name);
  
  return (l);
}

/* Find a location by ip. */

static AuthLocName *find_loc_by_ips(loc_dir_config_rec *dcfg, char *ips)
{
  AuthLocName *l;
  AuthLocIpRange *ip;
  PRINTF2("al: look for loc %s\n", ips);
  for (l=dcfg->loc; l; l=l->next) {
     for (ip=l->ip; ip; ip=ip->next) {
       PRINTF2("   ip: %d %s %d-%d\n", ip->type, ip->ipstr, ip->lo, ip->hi);
       switch (ip->type) {
          case IP_IP:
               if (!strcmp(ip->ipstr, ips)) return (l);
               break;
          case IP_STAR:
               if (!strncmp(ip->ipstr, ips, strlen(ip->ipstr))) return (l);
               break;
          case IP_RANGE:
               if (!strncmp(ip->ipstr, ips, strlen(ip->ipstr))) {
                  int ipn = atoi(ips+strlen(ip->ipstr));
                  PRINTF2("       tgt: %d\n", ipn);
                  if (ipn>=ip->lo || ipn<=ip->hi) return (l);
               }
               break;
       }
       PRINTF2("   not yet\n");
     }
  }
  PRINTF2("al: not found\n");
  return (NULL);
}

/* Find an app by name, Possibly make new if not found */

static AuthLocAppName *find_app_by_name(pool *p, loc_dir_config_rec *dir_conf,
     const char *name, int new)
{
  AuthLocAppName *last = NULL;
  AuthLocAppName *a;
  PRINTF2("al: look (%d) for app %s\n", new, name);
  for (a=dir_conf->app; a; last=a,a=a->next) {
     if (!strcmp(a->name, name)) return (a);
  }
  /* not found */
  PRINTF2("al: not found\n");
  if (!new) return (NULL);
  a = (AuthLocAppName*) ap_pcalloc (p, sizeof (AuthLocAppName));
  if (last) last->next = a;
  else dir_conf->app = a;
  a->name = ap_pstrdup(p, name);
  
  return (a);
}

/* Find an app by request string. */

static AuthLocAppName *find_app_by_str(loc_dir_config_rec *dcfg,
     char *uri, char *arg)
{
  AuthLocAppName *a;
  PRINTF2("al: look for app %s (%s)\n", uri, arg);
  for (a=dcfg->app; a; a=a->next) {
     if (a->where==AppURL && uri && strstr(uri, a->str)) return (a); 
     if (a->where==AppQS && arg && strstr(arg, a->str)) return (a); 
  }
  PRINTF2("al: not found\n");
  return (NULL);
}

/* Find a permit by name. Possibly make new if not found */

static AuthLocPermit *find_permit_by_name(pool *p, loc_dir_config_rec *dir_conf,
     const char *appname, const char *locname, int new)
{
  AuthLocPermit *last = NULL;
  AuthLocPermit *m;

  PRINTF2("al: look (%d) for pmt %s/%s\n", new, appname, locname);
  for (m=dir_conf->permit; m; last=m,m=m->next) {
     if (strcmp(m->appname, appname)) continue;
     if (strcmp(m->locname, locname)) continue;
     return (m);
  }
  PRINTF2("al: not found\n");
  /* not found */
  if (!new) return (NULL);
  m = (AuthLocPermit*) ap_pcalloc (p, sizeof (AuthLocPermit));
  if (last) last->next = m;
  else dir_conf->permit = m;
  
  m->appname = ap_pstrdup(p, appname);
  m->locname = ap_pstrdup(p, locname);

  m->type = 0;
  m->id = NULL;
  m->page = NULL;
  
  return (m);
}

/* Display a dir config.  Strictly a debugging tool.  */

static void 
show_loc_dir_config (loc_dir_config_rec *cfg, char *label)
{
  int i;
  AuthLocName *l;
  AuthLocAppName *a;
  AuthLocPermit *m;
  AuthLocIpRange *ip;

  if (1) return;

  PRINTF("<dir_config %s>\n", label);


     /* locations */
     for (l=cfg->loc; l; l=l->next) {
        PRINTF(" loc: %s\n", l->name);
        for (ip=l->ip; ip; ip=ip->next) {
           PRINTF("  ip: %d %s %d-%d\n", ip->type, ip->ipstr, ip->lo, ip->hi);
        }
     }

     /* applications */
     for (a=cfg->app; a; a=a->next) {
        PRINTF(" app: %s str=%s loc=%d\n", a->name, a->str, a->where);
     }

     /* permits */
     for (m=cfg->permit; m; m=m->next) {
        PRINTF(" pmt: %d app=%s loc=%s id=%s page=%s\n",
            m->type, m->appname, m->locname, m->id, m->page);
     }

}

/* Copy a "guest" form, substituting our keywords. */

static struct {
  char *key;
  char *rsp;
} page_key[] = {
 { PMT_PAGE_YES, PMT_RESP_YES },
 { PMT_PAGE_NO, PMT_RESP_NO }
};

static int copy_guest_page(request_rec *r, AuthLocPermit *pmt)
{
  FILE *fp;
  int fl;
  char *pag = NULL;
  int i;
  char *rp;
  char *kp[2];
  int ko[2];
  PRINTF("al: cpy guest page, %s\n", pmt->page?pmt->page:"null");

  /* Use configured page if available - else default */
  if (pmt->page&&(*pmt->page)&&(fp=fopen(pmt->page,"r"))) {
     fseek(fp, 0, SEEK_END);
     fl = ftell(fp);
     PRINTF2("al: guest len = %d\n", fl);
     if (fl>0) {
        fseek(fp,0,SEEK_SET);
        pag = (char*) malloc(fl+1);
        fread(pag, fl, 1, fp);
        pag[fl] = '\0';
     }
     fclose(fp);
  }
  if (!pag) {
    /* simple page */
    pag = strdup("<html><body><a href=\"{GUEST_YES}\">Be guest</a><p><a href=\"{GUEST_NO}\">Be yourself</a></html>\n");
    fl = strlen(pag);
  }
        
  /* find the replacement keys */
  kp[0] = strstr(pag, page_key[0].key);
  kp[1] = strstr(pag, page_key[1].key);
  PRINTF2("al: 0,1 at %d,%d\n", kp[0]-pag, kp[1]-pag);
  if (kp[0]<kp[1]) {
     ko[0] = 0;
     ko[1] = 1;
  } else {
     ko[0] = 1;
     ko[1] = 0;
  }
  /* Send page with link substitution */
  rp = pag;
  for (i=0; i<2; i++) {
     int k = ko[i];
     if (!kp[k]) continue; 
     ap_rwrite(rp, kp[k]-rp, r);
     ap_rputs(r->uri, r);
     if (r->args) {
        ap_rputs("?", r);
        ap_rputs(r->args, r);
        ap_rputs("&", r);
     } else {
        ap_rputs("?", r);
     }
     ap_rputs(page_key[k].rsp, r);
     rp = kp[k] + strlen(page_key[k].key);
  }
  ap_rwrite(rp, pag+fl-rp-1, r);
  free(pag);
}


/* ------- Configuration processing ----------- */

static void *
create_loc_dir_config (pool * p, char *d)
{
  loc_dir_config_rec *cfg = (loc_dir_config_rec *)
      ap_pcalloc (p, sizeof (loc_dir_config_rec));

  PRINTF2("<create_loc_dir_config> %s\n", d?d:"-root-");
  cfg->loc = NULL;
  cfg->app = NULL;
  cfg->permit = NULL;
  return cfg;
}


/* Dir merge */

static void *
merge_loc_dir_config (pool *p, void *basev, void *supplv)
{
  loc_dir_config_rec *new;
  loc_dir_config_rec *rec[2];
  int i;
  AuthLocName *l, *nl;
  AuthLocAppName *a, *na;
  AuthLocPermit *m, *nm;

  PRINTF2("<merge_loc_dir_config>\n");

  new = (loc_dir_config_rec *) create_loc_dir_config(p, "-merge-");

  rec[0] = (loc_dir_config_rec *) basev;
  rec[1] = (loc_dir_config_rec *) supplv;

  show_loc_dir_config(rec[0], "Base");
  show_loc_dir_config(rec[1], "Supl");

  for (i=0;i<2;i++) {
     AuthLocIpRange *ip, **ipp, *oip;

     /* copy locations */
     for (l=rec[i]->loc; l; l=l->next) {
        nl = find_loc_by_name(p, new, l->name, 1);
        for (ipp=&(nl->ip); *ipp; ipp=&((*ipp)->next));
        for (oip=l->ip; oip; oip=oip->next) {
           ip = (AuthLocIpRange*) malloc(sizeof(AuthLocIpRange));
           ip->next = NULL;
           ip->ipstr = ap_pstrdup(p, oip->ipstr);
           ip->type = oip->type;
           ip->lo = oip->lo;
           ip->hi = oip->hi;
           *ipp = ip;
           ipp = &(ip->next);
        }
     }

     /* copy applications */
     for (a=rec[i]->app; a; a=a->next) {
        na = find_app_by_name(p, new, a->name, 1);
        na->str = ap_pstrdup(p, a->str);
        na->where = a->where;
     }

     /* copy permits */
     for (m=rec[i]->permit; m; m=m->next) {
        nm = find_permit_by_name(p, new, m->appname, m->locname, 1);
        nm->id = ap_pstrdup(p, m->id);
        nm->page = ap_pstrdup(p, m->page);
        nm->type = m->type;
     }
  }

  show_loc_dir_config(new, "Merged");

  return new;
}


/* ------- Directive processing ----------- */

/* Process the "AuthLocationDefineLocation" directive */

const char *set_loc_define_location (cmd_parms *cmd,
      void *dir_confv, const char *arg)
{
    void *server_conf = cmd->server->module_config;
    loc_dir_config_rec *dir_conf = (loc_dir_config_rec*) dir_confv;

    char *w;
    AuthLocName *loc, **locp;
    char *vargs;
    pool *p = cmd->pool;
    AuthLocIpRange *ip;

    PRINTF2("<set_loc_define_location>\n");

    vargs = ap_pstrdup(p, arg);
    loc = find_loc_by_name(p, dir_conf, next_word(&vargs), 1);

    PRINTF2(" .. name = %s \n", loc->name);

    /* add each ip entry */
    while ((w=next_word(&vargs)) && *w) {
       AuthLocIpRange *ip, **ipp;
       ip = parse_ip(p, w);
       PRINTF2(" ...  ip = %d: %s (%d/%d)\n", ip->type, ip->ipstr,
               ip->lo, ip->hi);
       for (ipp=&(loc->ip); *ipp; ipp=&((*ipp)->next));
       *ipp = ip;
    }
    return (NULL);

}



/* Process the "AuthLocationDefineApplication" directive */

const char *set_loc_define_application (cmd_parms *cmd,
      void *dir_confv, const char *arg1, const char *arg2, const char *arg3)
{
    void *server_conf = cmd->server->module_config;
    loc_dir_config_rec *dir_conf = (loc_dir_config_rec*) dir_confv;

    char *w;
    AuthLocAppName *app, *appp;
    pool *p = cmd->pool;

    PRINTF2("<set_loc_define_app>\n");

    app = find_app_by_name(p, dir_conf, arg1, 1);

    PRINTF2(" .. name = %s \n", app->name);

    app->str = ap_pstrdup(p, arg2);

    if (!strcmp(arg3, "URL")) app->where = AppURL;
    else if (!strcmp(arg3, "QS")) app->where = AppQS;
    else return "bad spec";

    return (NULL);

}


/* Process the "AuthLocationPermitAnonymous" directive */

const char *set_loc_permit_anon (cmd_parms *cmd,
      void *dir_confv, const char *arg)
{
    void *server_conf = cmd->server->module_config;
    char *appname, *locname;
    loc_dir_config_rec *dir_conf = (loc_dir_config_rec*) dir_confv;

    AuthLocPermit *pmt;
    char *vargs;
    pool *p = cmd->pool;

    PRINTF2("<set_loc_permit_anon>\n");

    vargs = ap_pstrdup(p, arg);

    appname = next_word(&vargs);
    locname = next_word(&vargs);

    pmt = find_permit_by_name(p, dir_conf, appname, locname, 1);
    if (!pmt) return ("No such app or loc");
    pmt->type = PERMIT_PMT_ANON;
    pmt->id = next_word(&vargs);
    pmt->page = next_word(&vargs);
    pmt->type = PERMIT_PMT_ANON;

    return (NULL);

}


/* Process the "AuthLocationRequireAnonymous" directive */

const char *set_loc_require_anon (cmd_parms *cmd,
      void *dir_confv, const char *arg1, const char *arg2, const char *arg3)
{
    void *server_conf = cmd->server->module_config;
    loc_dir_config_rec *dir_conf = (loc_dir_config_rec*) dir_confv;

    AuthLocPermit *pmt;
    char *vargs;
    pool *p = cmd->pool;

    PRINTF2("<set_loc_require_anon>\n");

    pmt = find_permit_by_name(p, dir_conf, arg1, arg2, 1);
    pmt->id = ap_pstrdup(p, arg3);
    pmt->type = PERMIT_REQ_ANON;

    return (NULL);

}


/* ----- Request processing -------- */


/* To avoid competing with other auth modules for priority
   of request handling we work in the "access" step,
   just prior to the auth atages. If we decide to set this
   to a guest id, we alsp set the type to "Location". */


static int loc_check_access(request_rec *r)
{
   loc_dir_config_rec *dcfg = (loc_dir_config_rec*)
      ap_get_module_config(r->per_dir_config, &auth_location_module);

   AuthLocName *loc;
   AuthLocAppName *app = NULL;
   AuthLocPermit *pmt;
   int ret = DECLINED;
   core_dir_config *ccfg;

   PRINTF("al: chk acc: sub=%s, uri='%s', arg='%s'\n", 
         r->main?"sub":"main", r->uri, r->args);
   /* don't bother with sub-requests */
   if (r->main) return(DECLINED);

   loc = find_loc_by_ips(dcfg, r->connection->remote_ip);
   if (!loc) return (DECLINED);
   PRINTF("al: loc = %s\n", loc?loc->name:"(null)");
   
   app = find_app_by_str(dcfg, r->uri, r->args);
   if (!app) return (DECLINED);
   PRINTF("al: app = %s\n", app->name);
   
  
   /* Check the permit directives */

   if (pmt = find_permit_by_name(r->pool, dcfg, app->name, loc->name, 0)) {
      char *rs;
      switch (pmt->type) {

         case PERMIT_PMT_ANON:

           if (r->args && (rs=strstr(r->args, PMT_RESP_YES))) {
              PRINTF("al: set permit anon id=%s\n", pmt->id);
              ap_log_rerror(MY_LOG_INFO, r,
                   "mod_auth_loc: setting user to s\n", pmt->id);
              r->USER = ap_pstrdup(r->pool, pmt->id);
              ccfg = (core_dir_config *)ap_get_module_config(r->per_dir_config,
                                         &core_module);
              ccfg->ap_auth_type = "Location";
              r->AUTH_TYPE = "Location";
           } else if (r->args && (rs=strstr(r->args, PMT_RESP_NO))) {
              PRINTF("al: user says no guest\n");
              ap_log_rerror(MY_LOG_INFO, r,
                   "mod_auth_loc: user says no guest\n");
           } else {
              r->content_type = "text/html";
              ap_send_http_header(r);
              copy_guest_page(r, pmt);
              ret = DONE;
           }
           break;

         case PERMIT_REQ_ANON:

           PRINTF("al: set require anon id=%s\n", pmt->id);
           ap_log_rerror(MY_LOG_INFO, r,
                   "mod_auth_loc: setting user to s\n", pmt->id);
           r->USER = ap_pstrdup(r->pool, pmt->id);
           ccfg = (core_dir_config *)ap_get_module_config(r->per_dir_config,
                                         &core_module);
           ccfg->ap_auth_type = "Location";
           r->AUTH_TYPE = "Location"; 
           break;

         default:
           break;
      }
   }

   return (ret);
}


/* Id check: If we issued the id, say OK */

static int loc_check_user(request_rec *r)
{
   loc_dir_config_rec *dcfg = (loc_dir_config_rec*)
      ap_get_module_config(r->per_dir_config, &auth_location_module);

   PRINTF("al: check_user, id=%s, type=%s(%s)\n", r->USER,
        ap_auth_type(r), r->AUTH_TYPE);
   if (!strcmp(ap_auth_type(r),"Location")) {
      PRINTF("al: my auth user\n");
      return (OK);
   }
   return (DECLINED);
}

/* Auth check: We don't actually do any auth */

static int loc_check_auth(request_rec *r)
{
   loc_dir_config_rec *dcfg = (loc_dir_config_rec*)
      ap_get_module_config(r->per_dir_config, &auth_location_module);

   return (DECLINED);
}

/* Fixup: Remove and arg string we might hae added */

static int loc_fixup(request_rec *r)
{
   loc_dir_config_rec *dcfg = (loc_dir_config_rec*)
      ap_get_module_config(r->per_dir_config, &auth_location_module);
   char *s, *a;

/* Don't remove the selection arg.  It might be needed on a redirect,
   as, for example, by autoindex. We could make this optonal by directive. */
   if (1) return (DECLINED);

   PRINTF("al: fixup\n");
   if (r->args) {
      if ((s=strstr(r->args,PMT_RESP_YES))||(s=strstr(r->args,PMT_RESP_NO))) {
         a = strchr(s, '&');
         if (s>r->args && *(s-1)=='&') s--;
         if (a) memmove(s, a+1, strlen(a+1)+1);
         else *s = '\0';
         if (!*r->args) r->args = NULL;
      }
   }
   return (DECLINED);
}



/* ------  Configuration ---------- */

#ifdef APACHE1_3
#define AP_INIT_TAKE1(d,f,c,w,h) { d,f,c,w,TAKE1,h}
#define AP_INIT_TAKE3(d,f,c,w,h) { d,f,c,w,TAKE3,h}
#define AP_INIT_NO_ARGS(d,f,c,w,h) { d,f,c,w,NO_ARGS,h}
#define AP_INIT_RAW_ARGS(d,f,c,w,h) { d,f,c,w,RAW_ARGS,h}
#define AP_INIT_ITERATE(d,f,c,w,h) { d,f,c,w,ITERATE,h}
#define AP_INIT_FLAG(d,f,c,w,h) { d,f,c,w,FLAG,h}
#define AP_INIT_ITERATE2(d,f,c,w,h) { d,f,c,w,ITERATE2,h}
#endif


static const command_rec loc_cmds[] =
{
  /* server directives */
  AP_INIT_RAW_ARGS("AuthLocationDefineLocation",
      set_loc_define_location, NULL, RSRC_CONF|OR_AUTHCFG,
     "[define location name]"
  ),
     
  AP_INIT_TAKE3("AuthLocationDefineApplication",
      set_loc_define_application, NULL, RSRC_CONF|OR_AUTHCFG,
     "[define application name]"
  ),
     
  AP_INIT_RAW_ARGS("AuthLocationPermitAnonymous",
      set_loc_permit_anon, NULL, RSRC_CONF|OR_AUTHCFG,
     "[permit anonymous login]"
  ),
     
  AP_INIT_TAKE3("AuthLocationRequireAnonymous",
      set_loc_require_anon, NULL, RSRC_CONF|OR_AUTHCFG,
     "[require anonymous login]"
  ),
     
  {NULL}
};

#ifdef APACHE1_3

module auth_location_module = {
  STANDARD_MODULE_STUFF,
  NULL,				/* initializer */
  create_loc_dir_config,   	/* dir config creater */
  merge_loc_dir_config,		/* dir merger */
  NULL,	                        /* server config */
  NULL,	                        /* merge server configs */
  loc_cmds,			/* command table */
  NULL,				/* handlers */
  NULL,				/* filename translation */
  loc_check_user,		/* check_user_id */
  loc_check_auth,		/* check auth */
  loc_check_access,		/* check access */
  NULL,				/* type_checker */
  loc_fixup,			/* fixups */
  NULL,				/* logger */
  NULL,                         /* [3] header parser */
  NULL,                         /* process initializer */
  NULL,                         /* process exit/cleanup */
  NULL                          /* [1] post read_request handling */
};

#else /* apache 2.x */

static void register_hooks(pool *p) {
    ap_hook_access_checker(loc_check_access, NULL, NULL, APR_HOOK_FIRST);
    ap_hook_check_user_id(loc_check_user, NULL, NULL, APR_HOOK_FIRST);
    ap_hook_auth_checker(loc_check_auth, NULL, NULL, APR_HOOK_FIRST);
    ap_hook_fixups(loc_fixup, NULL, NULL, APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA auth_location_module = {
    STANDARD20_MODULE_STUFF,
    create_loc_dir_config,
    merge_loc_dir_config,
    NULL,
    NULL,
    loc_cmds,
    register_hooks,
};
#endif /* which apache */




